#!/usr/bin/perl
use strict;
use warnings;
use Tk;
use Tk::FileSelect;

use constant SPEED => 150; # milliseconds between each turn

use constant X => 0;
use constant Y => 1;
use constant POS => 0;
use constant PACKAGES => 1;
use constant STATE => 2;
use constant OLDPOS => 3;

my $tile_size = 32;

# Set up the Tk stuff
my $main_window = MainWindow->new(
  -title => 'Robot Replay',
);
my $canvas = $main_window->Scrolled(
  'Canvas',
  -scrollbars => 'se',
  -xscrollincrement => $tile_size,
  -yscrollincrement => $tile_size,
  -width => 21*$tile_size,
  -height => 20*$tile_size, # make it fit the 20x21 maps at first
);
$canvas->pack(
  -expand => 1,
  -fill => 'both',
);

use FindBin;

my %images;
$images{water} = $main_window->Photo(-file => "$FindBin::Bin/water.gif");
$images{wall} = $main_window->Photo(-file => "$FindBin::Bin/wall.gif");
$images{base} = $main_window->Photo(-file => "$FindBin::Bin/base.gif");
$images{path} = $main_window->Photo(-file => "$FindBin::Bin/path.gif");
$images{robot} = $main_window->Photo(-file => "$FindBin::Bin/robot.gif");
$images{robotpack} = $main_window->Photo(-file => "$FindBin::Bin/robot_pack.gif");
$images{robotdead} = $main_window->Photo(-file => "$FindBin::Bin/robot_dead.gif");

my @robot_offset = (-4, -60);

my %tile_types = ('#' => 'wall', '~' => 'water', '@' => 'base', '.' => 'path');

my (@board, @board_size, @moves);
my $robots = [];
my @map_tiles;
my @bot_images;

run_log();
MainLoop;

sub run_log {
  open_file();
  
  $canvas->delete($_) foreach @map_tiles; @map_tiles = ();
  $canvas->delete($_) foreach @bot_images; @bot_images = ();
  $robots = [];

  display_board();
  
  perform_turn();
}

sub display_board {
  $canvas->configure(
    -scrollregion => [0, 0, $board_size[0]*$tile_size, $board_size[1]*$tile_size],
  );
  for my $y (1..$board_size[1]) {
    for my $x (1..$board_size[0]) {
      my $char = substr($board[$board_size[1]-$y+1], $x-1, 1);
      push @map_tiles, $canvas->createImage(($x-1)*$tile_size,
                                            ($y-1)*$tile_size,
                                            -anchor => 'nw',
                                            -image => $images{$tile_types{$char}});
    }
  }
}

sub open_file {

  # getOpenFile seems to be buggy and blocks all input...
#  my $filename = $canvas->getOpenFile(-title => 'Select a log file to watch');

  my $FS = $main_window->FileSelect();
  my $filename = $FS->Show();

  exit unless $filename;

  open LOG, "$filename" or die "Error opening $filename for input: $!";

  chomp (my $board_sizes = <LOG>);
  die "Invalid log file header!" unless $board_sizes =~ /^\d+\s+\d+$/;
  @board_size = split /\s+/, $board_sizes;

  @board = ();
  for (1..$board_size[1]) {
    chomp (my $line = <LOG>);
    $board[$_] = $line;
  }

  @moves = <LOG>;
  close LOG;
}

sub perform_turn {

  parse_response(shift @moves);

  if (@moves) {
    $main_window->after(SPEED, \&perform_turn);
  } else {
    my $response = $main_window->messageBox(
      -message => 'Finished! View another log file?',
      -type => 'YesNo',
      -title => 'Status'
    );
    if ($response eq 'yes') {
       (@board, @board_size, @moves) = ();
       $robots = [];
       @map_tiles = ();
       @bot_images = ();
       run_log;
    } else {
      exit;
    }
  }
}


sub parse_response {
  my $used_robots = [];
  my $response_data = shift;

  my @responses = split(/\s+/, $response_data);
  my $robot_id = -1;

  while (defined (my $command = shift @responses)) {

    if (substr($command, 0, 1) eq '#') {
      $robot_id = substr($command, 1);
      ++$used_robots->[$robot_id];

    } elsif ($command eq 'N') {
      $robots->[$robot_id][POS][Y] = $robots->[$robot_id][POS][Y]+1;
      $canvas->move($bot_images[$robot_id], 0, -$tile_size);

    } elsif ($command eq 'S') {
      $robots->[$robot_id][POS][Y] = $robots->[$robot_id][POS][Y]-1;
      $canvas->move($bot_images[$robot_id], 0, $tile_size);

    } elsif ($command eq 'E') {
      $robots->[$robot_id][POS][X] = $robots->[$robot_id][POS][X]+1;
      $canvas->move($bot_images[$robot_id], $tile_size, 0);

    } elsif ($command eq 'W') {
      $robots->[$robot_id][POS][X] = $robots->[$robot_id][POS][X]-1;
      $canvas->move($bot_images[$robot_id], -$tile_size, 0);

    } elsif ($command eq 'X') {
      $robots->[$robot_id][POS][X] = shift @responses;
      shift @responses;
      $robots->[$robot_id][POS][Y] = shift @responses;

      $canvas->delete($bot_images[$robot_id]) if $bot_images[$robot_id]; # delete any existing ones

      $bot_images[$robot_id] = $canvas->createImage(
        ($robots->[$robot_id][POS][X]-1) * $tile_size + $robot_offset[X],
        ($board_size[1]-$robots->[$robot_id][POS][Y]) * $tile_size + $robot_offset[Y],
        -anchor => 'nw',
        -image => $images{robot}
      );

    } elsif ($command eq 'P') {
      $robots->[$robot_id][STATE] = ['picking'];
      my $package_id = shift @responses;
      push @{$robots->[$robot_id][PACKAGES]}, $package_id;
      $canvas->itemconfigure($bot_images[$robot_id], -image => $images{robotpack});

    } elsif ($command eq 'D') {
      $robots->[$robot_id][STATE] = ['dropping'];
      my $package_id = shift @responses;
      @{$robots->[$robot_id][PACKAGES]} = grep $_ != $package_id, @{$robots->[$robot_id][PACKAGES]};
      if (not @{$robots->[$robot_id][PACKAGES]}) {
        $canvas->itemconfigure($bot_images[$robot_id], -image => $images{robot});
      }

    } else {
      print "Unrecognised command $command!\n";
    }
  }

  foreach my $id (0..$#$robots) {
    if (defined $robots->[$id] and not $used_robots->[$id]) {
      $canvas->itemconfigure($bot_images[$id], -image => $images{robotdead});
    }
  }
}
